@@ -6,9 +6,13 @@ import uuid
 
 from deprecated import deprecated
 
-from jwcrypto.common import JWException, json_decode, json_encode
+from jwcrypto.common import JWException, JWKeyNotFound
+from jwcrypto.common import json_decode, json_encode
 from jwcrypto.jwe import JWE
+from jwcrypto.jwe import default_allowed_algs as jwe_algs
+from jwcrypto.jwk import JWK, JWKSet
 from jwcrypto.jws import JWS
+from jwcrypto.jws import default_allowed_algs as jws_algs
 
 
 # RFC 7519 - 4.1
@@ -20,6 +24,17 @@ JWTClaimsRegistry = {'iss': 'Issuer',
                      'nbf': 'Not Before',
                      'iat': 'Issued At',
                      'jti': 'JWT ID'}
+"""Registry of RFC 7519 defined claims"""
+
+
+# do not use this unless you know about CVE-2022-3102
+JWT_expect_type = True
+"""This module parameter can disable the use of the expectation
+   feature that has been introduced to fix CVE-2022-3102. This knob
+   has been added as a workaround for applications that can't be
+   immediately refactored to deal with the change in behavior but it
+   is considered deprecated and will be removed in a future release.
+"""
 
 
 class JWTExpired(JWException):
@@ -127,8 +142,7 @@ class JWTMissingKeyID(JWException):
         super(JWTMissingKeyID, self).__init__(msg)
 
 
-@deprecated
-class JWTMissingKey(JWException):
+class JWTMissingKey(JWKeyNotFound):
     """JSON Web Token is using a key not in the key set.
 
     This exception is raised if the key that was used is not available
@@ -153,7 +167,8 @@ class JWT:
     """
 
     def __init__(self, header=None, claims=None, jwt=None, key=None,
-                 algs=None, default_claims=None, check_claims=None):
+                 algs=None, default_claims=None, check_claims=None,
+                 expected_type=None):
         """Creates a JWT object.
 
         :param header: A dict or a JSON string with the JWT Header data.
@@ -169,6 +184,12 @@ class JWT:
         :param check_claims: An optional dict of claims that must be
          present in the token, if the value is not None the claim must
          match exactly.
+        :param expected_type: An optional string that defines what kind
+         of token to expect when validating a deserialized token.
+         Supported values: "JWS" or "JWE"
+         If left to None the code will try to detect what the expected
+         type is based on other parameters like 'algs' and will default
+         to JWS if no hints are found. It has no effect on token creation.
 
         Note: either the header,claims or jwt,key parameters should be
         provided as a deserialization operation (which occurs if the jwt
@@ -190,6 +211,7 @@ class JWT:
         self._leeway = 60  # 1 minute clock skew allowed
         self._validity = 600  # 10 minutes validity (up to 11 with leeway)
         self.deserializelog = None
+        self._expected_type = expected_type
 
         if header:
             self.header = header
@@ -276,6 +298,96 @@ class JWT:
     def validity(self, v):
         self._validity = int(v)
 
+    def _expected_type_heuristics(self, key=None):
+        if self._expected_type is None and self._algs:
+            if set(self._algs).issubset(jwe_algs + ['RSA1_5']):
+                self._expected_type = "JWE"
+            elif set(self._algs).issubset(jws_algs):
+                self._expected_type = "JWS"
+        if self._expected_type is None and self._header:
+            if "enc" in json_decode(self._header):
+                self._expected_type = "JWE"
+        if self._expected_type is None and key is not None:
+            if isinstance(key, JWK):
+                use = key.get('use')
+                if use == 'sig':
+                    self._expected_type = "JWS"
+                elif use == 'enc':
+                    self._expected_type = "JWE"
+            elif isinstance(key, JWKSet):
+                all_use = None
+                # we can infer only if all keys are of the same type
+                for k in key:
+                    use = k.get('use')
+                    if all_use is None:
+                        all_use = use
+                    elif use != all_use:
+                        all_use = None
+                        break
+                if all_use == 'sig':
+                    self._expected_type = "JWS"
+                elif all_use == 'enc':
+                    self._expected_type = "JWE"
+        if self._expected_type is None and key is not None:
+            if isinstance(key, JWK):
+                ops = key.get('key_ops')
+                if ops:
+                    if not isinstance(ops, list):
+                        ops = [ops]
+                    if set(ops).issubset(['sign', 'verify']):
+                        self._expected_type = "JWS"
+                    elif set(ops).issubset(['encrypt', 'decrypt']):
+                        self._expected_type = "JWE"
+            elif isinstance(key, JWKSet):
+                all_ops = None
+                ttype = None
+                # we can infer only if all keys are of the same type
+                for k in key:
+                    ops = k.get('key_ops')
+                    if ops:
+                        if not isinstance(ops, list):
+                            ops = [ops]
+                        if all_ops is None:
+                            if set(ops).issubset(['sign', 'verify']):
+                                all_ops = set(['sign', 'verify'])
+                                ttype = "JWS"
+                            elif set(ops).issubset(['encrypt', 'decrypt']):
+                                all_ops = set(['encrypt', 'decrypt'])
+                                ttype = "JWE"
+                            else:
+                                ttype = None
+                                break
+                        else:
+                            if not set(ops).issubset(all_ops):
+                                ttype = None
+                                break
+                    elif all_ops:
+                        ttype = None
+                        break
+                if ttype:
+                    self._expected_type = ttype
+        if self._expected_type is None:
+            self._expected_type = "JWS"
+        return self._expected_type
+
+    @property
+    def expected_type(self):
+        if self._expected_type is not None:
+            return self._expected_type
+
+        # If no expected type is set we default to accept only JWSs,
+        # however to improve backwards compatibility we try some
+        # heuristic to see if there has been strong indication of
+        # what the expected token type is.
+        return self._expected_type_heuristics()
+
+    @expected_type.setter
+    def expected_type(self, v):
+        if v in ["JWS", "JWE"]:
+            self._expected_type = v
+        else:
+            raise ValueError("Invalid value, must be 'JWS' or 'JWE'")
+
     def _add_optional_claim(self, name, claims):
         if name in claims:
             return
@@ -472,6 +584,7 @@ class JWT:
             t.allowed_algs = self._algs
         t.add_signature(key, protected=self.header)
         self.token = t
+        self._expected_type = "JWS"
 
     def make_encrypted_token(self, key):
         """Encrypts the payload.
@@ -488,6 +601,7 @@ class JWT:
             t.allowed_algs = self._algs
         t.add_recipient(key)
         self.token = t
+        self._expected_type = "JWE"
 
     def validate(self, key):
         """Validate a JWT token that was deserialized w/o providing a key
@@ -500,13 +614,22 @@ class JWT:
         if self.token is None:
             raise ValueError("Token empty")
 
+        et = self._expected_type_heuristics(key)
+        validate_fn = None
+
+        if isinstance(self.token, JWS):
+            if et != "JWS" and JWT_expect_type:
+                raise TypeError("Expected {}, got JWS".format(et))
+            validate_fn = self.token.verify
+        elif isinstance(self.token, JWE):
+            if et != "JWE" and JWT_expect_type:
+                raise TypeError("Expected {}, got JWE".format(et))
+            validate_fn = self.token.decrypt
+        else:
+            raise ValueError("Token format unrecognized")
+
         try:
-            if isinstance(self.token, JWS):
-                self.token.verify(key)
-            elif isinstance(self.token, JWE):
-                self.token.decrypt(key)
-            else:
-                raise ValueError("Token format unrecognized")
+            validate_fn(key)
             self.deserializelog.append("Success")
         except Exception as e:  # pylint: disable=broad-except
             if isinstance(self.token, JWS):
@@ -515,10 +638,15 @@ class JWT:
                 self.deserializelog = self.token.decryptlog
             self.deserializelog.append(
                 'Validation failed: [{}]'.format(repr(e)))
+            if isinstance(e, JWKeyNotFound):
+                raise JWTMissingKey() from e
             raise
 
         self.header = self.token.jose_header
-        self.claims = self.token.payload.decode('utf-8')
+        payload = self.token.payload
+        if isinstance(payload, bytes):
+            payload = payload.decode('utf-8')
+        self.claims = payload
         self._check_provided_claims()
 
     def deserialize(self, jwt, key=None):
